using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Battlehub.RTSaveLoad.PersistentObjects.UI;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Serialization;

namespace CheesyFX
{
    public class JiggleHandler
    {
        public string region;
        private int regionId;
        private int sideId = -1;
        private int subId;
        private RBDynamics rbDynamic;
        

        private bool[] isJiggling = new bool[4];
        private float[] amp = new float[4];
        private float[] phase = new float[4];

        public DAZMorph[] morphs;

        private Vector4 strength { get; set; }
        private Vector4 frequency { get; set; }
        private Vector4 damping { get; set; }
        private Vector4 minAccel { get; set; }
        private Vector4 maxAccel { get; set; }
        private Vector4 maxMorph { get; set; }
        
        private Vector4 strengthTense { get; set; }
        private Vector4 frequencyTense { get; set; }
        private Vector4 dampingTense { get; set; }
        
        public bool tensionEnabled { get; set; }

        public JSONStorableVector4 strengthJSON = new JSONStorableVector4(
            "strength",
            new Vector4(3f, 2f, 3f, 2f),
            new Vector4(),
            new Vector4(10f,10f,10f, 10f));
        
        public JSONStorableVector4 strengthTenseJSON = new JSONStorableVector4(
            "strTense",
            new Vector4(2f, 1f, 2f, 2f),
            new Vector4(),
            new Vector4(10f,10f,10f, 10f));

        public JSONStorableVector4 frequencyJSON = new JSONStorableVector4(
            "frequency",
            new Vector4(16f, 16, 16f,16f),
            new Vector4(),
            new Vector4(25f,25f,25f,25f));
        
        public JSONStorableVector4 frequencyTenseJSON = new JSONStorableVector4(
            "freqTense",
            new Vector4(22f, 22, 22f,22f),
            new Vector4(),
            new Vector4(25f,25f,25f,25f));
        
        public JSONStorableVector4 dampingJSON = new JSONStorableVector4(
            "damping",
            new Vector4(5f, 5f, 5f,5f),
            new Vector4(),
            new Vector4(10f,10f,10f,10f));
        
        public JSONStorableVector4 dampingTenseJSON = new JSONStorableVector4(
            "dampTense",
            new Vector4(7f, 7f, 7f,7f),
            new Vector4(),
            new Vector4(10f,10f,10f,10f));
        
        public JSONStorableVector4 minAccelJSON = new JSONStorableVector4(
            "minAccel",
            new Vector4(2f, 2f, 2f,2f),
            new Vector4(),
            new Vector4(5f,5f,5f,5f));
        
        public JSONStorableVector4 maxAccelJSON = new JSONStorableVector4(
            "maxAccel",
            new Vector4(3f, 3f, 3f,2.2f),
            new Vector4(),
            new Vector4(5f,5f,5f,5f));
        
        public JSONStorableVector4 maxMorphJSON = new JSONStorableVector4(
            "maxMorph",
            new Vector4(1f, 1f, 1f, 1f),
            new Vector4(),
            new Vector4(3f,3f,3f,3f));
        
        // Monitoring
        public JSONStorableVector4 strengthMonitorJSON = new JSONStorableVector4(
            "strength",
            new Vector4(3f, 2f, 3f, 2.5f),
            new Vector4(),
            new Vector4(10f,10f,10f, 10f));
        
        public JSONStorableVector4 frequencyMonitorJSON = new JSONStorableVector4(
            "freqMon",
            new Vector4(16f, 16, 16f,16f),
            new Vector4(),
            new Vector4(25f,25f,25f,25f));
        
        public JSONStorableVector4 dampingMonitorJSON = new JSONStorableVector4(
            "damping",
            new Vector4(5f, 5f, 5f,5f),
            new Vector4(),
            new Vector4(10f,10f,10f,10f));

        private const float TwoPi = 2f*Mathf.PI;
        private const float PhaseToId = 360f/(2f*Mathf.PI);
        private float _directionFactor = 1f;

        public JiggleHandler(int id, string region, RBDynamics rbDynamic)
        {
            this.region = region;
            regionId = id;
            if (region.StartsWith("l")) sideId = 0;
            else if (region.StartsWith("r")) sideId = 1;
            if (region.Contains("Top")) subId = 0;
            else if (region.Contains("Top")) subId = 1;
            else if (region.Contains("Top")) subId = 2;
            this.rbDynamic = rbDynamic;
            tensionEnabled = true;
            if (region.Contains("UArm"))
            {
                morphs = new[]
                {
                    ShakeIt.GetMorphByUid($"SI_{region}_Y"),
                    ShakeIt.GetMorphByUid($"SI_{region}_X"),
                    ShakeIt.GetMorphByUid($"SI_{region}_Z"),
                    ShakeIt.GetMorphByUid($"SI_{region}_R"),
                };
            }
            else
            {
                morphs = new[]
                {
                    ShakeIt.GetMorphByUid($"SI_{region}_X"),
                    ShakeIt.GetMorphByUid($"SI_{region}_Y"),
                    ShakeIt.GetMorphByUid($"SI_{region}_Z"),
                    ShakeIt.GetMorphByUid($"SI_{region}_R"),
                };
            }
            ApplyDefaults();
            strengthJSON.setCallbackFunction += val => strength = val * .01f;
            strengthTenseJSON.setCallbackFunction += val => strengthTense = val * .01f;
            frequencyJSON.setCallbackFunction += val => frequency = val;
            frequencyTenseJSON.setCallbackFunction += val => frequencyTense = val;
            dampingJSON.setCallbackFunction += val => damping = val;
            dampingTenseJSON.setCallbackFunction += val => dampingTense = val;
            minAccelJSON.setCallbackFunction += val => minAccel = val;
            maxAccelJSON.setCallbackFunction += val => maxAccel = val;
            maxMorphJSON.setCallbackFunction += val => maxMorph = val;
        }

        public void ApplyDefaults()
        {
            var strengthDefault = strengthJSON.defaultVal;
            var strengthTenseDefault = strengthTenseJSON.defaultVal;
            var frequencyDefault = frequencyJSON.defaultVal;
            var frequencyTenseDefault = frequencyTenseJSON.defaultVal;
            var dampingDefault = dampingJSON.defaultVal;
            var dampingTenseDefault = dampingTenseJSON.defaultVal;
            var minAccelDefault = minAccelJSON.defaultVal;
            var maxAccelDefault = maxAccelJSON.defaultVal;
            var maxMorphDefault = maxMorphJSON.defaultVal;
            if (region.Contains("Calf"))
            {
                strengthDefault.z = 2f;
                strengthTenseDefault.z = 1f;
                dampingDefault = new Vector4(4f, 4f, 3f, 4f);
                
            }
            if(region.Contains("Abs"))
            {
                strengthDefault = new Vector4(1f, 2f, 2f, 1f);
                frequencyDefault = new Vector4(18f, 18f, 18f, 18f);
            }

            strengthJSON.defaultVal = strengthDefault;
            strengthJSON.val = strengthDefault;
            strength = strengthDefault * .01f;
            strengthTenseJSON.defaultVal = strengthTenseDefault;
            strengthTenseJSON.val = strengthTenseDefault;
            strengthTense = strengthTenseDefault * .01f;
            
            frequencyJSON.defaultVal = frequencyDefault;
            frequencyJSON.val = frequencyDefault;
            frequency = frequencyDefault;
            frequencyTenseJSON.defaultVal = frequencyTenseDefault;
            frequencyTenseJSON.val = frequencyTenseDefault;
            frequencyTense = frequencyTenseDefault;
            
            dampingJSON.defaultVal = dampingDefault;
            dampingJSON.val = dampingDefault;
            damping = dampingDefault;
            dampingTenseJSON.defaultVal = dampingTenseDefault;
            dampingTenseJSON.val = dampingTenseDefault;
            dampingTense = dampingTenseDefault;
            
            minAccelJSON.defaultVal = minAccelDefault;
            minAccelJSON.val = minAccelDefault;
            minAccel = minAccelDefault;
            
            maxAccelJSON.defaultVal = maxAccelDefault;
            maxAccelJSON.val = maxAccelDefault;
            maxAccel = maxAccelDefault;
            
            maxMorphJSON.defaultVal = maxMorphDefault;
            maxMorphJSON.val = maxMorphDefault;
            maxMorph = maxMorphDefault;
        }

        public void Reset()
        {
            for (int i=0; i<morphs.Length; i++)
            {
                if(morphs[i] == null) continue;
                morphs[i].Reset();
            }
        }

        public void OnDisable()
        {
            Reset();
        }

        public void Jiggle()
        {
            float mag = 0f;
            Vector3 acc = rbDynamic.localAcceleration;
            
            for(int i = 0; i<4; i++)
            {
                if(morphs[i] == null) continue;
                switch (i)
                {
                    case 0:
                    {
                        mag = acc.x * _directionFactor;
                        break;
                    }
                    case 1:
                    {
                        mag = acc.y;
                        break;
                    }
                    case 2:
                    {
                        mag = acc.z;
                        break;
                    }
                    case 3:
                    {
                        mag = Mathf.Sqrt(acc.x * acc.x + acc.z * acc.z);
                        break;
                    }
                }
                if (Mathf.Abs(mag) > maxAccel[i]) mag = maxAccel[i] * Mathf.Sign(mag);
                JiggleDir(i, mag);
            }
        }

        private void JiggleDir(int dir, float mag)
        {
            var freq = frequency[dir];
            var damp = damping[dir];
            if (tensionEnabled)
            {
                var tension = rbDynamic.tension;
                freq = Mathf.Lerp(freq, frequencyTense[dir], tension);
                damp = Mathf.Lerp(damp, dampingTense[dir], tension);
                if(ShakeIt.monitor.val && regionId == ShakeIt.selectedRegionId && sideId <= 0)
                {
                    frequencyMonitorJSON.val = frequencyMonitorJSON.val.SetComponent(dir, freq);
                }
            }
            
            float a = amp[dir];
            
            if (Math.Abs(mag) > minAccel[dir])
            {
                mag *= strength[dir];
                if (!isJiggling[dir])
                {
                    a = Clamp(mag, dir);
                    isJiggling[dir] = true;
                }
                else
                {
                    float newAmp = Clamp(a + Mathf.Abs(mag)*Mathf.Sign(a), dir);
                    a = newAmp;
                }
            }
            
            if (Mathf.Abs(a) > .005f)
            {
                
                a = Mathf.Lerp(a, 0f, damp*Time.fixedUnscaledDeltaTime);
                float p = phase[dir];
                p += freq  * Time.fixedUnscaledDeltaTime;
                if (p > TwoPi) p = 0f;
                float _sin = a * ShakeIt.sin[(int)(p * PhaseToId)];
                morphs[dir].morphValue = _sin;
                amp[dir] = a;
                phase[dir] = p;
            }
            else if (isJiggling[dir])
            {
                phase[dir] = 0f;
                isJiggling[dir] = false;
            }
        }
        
        private float Clamp(float val, int dir)
        {
            var mVal = maxMorph[dir];
            if (val > mVal) return mVal;
            if (val < -mVal) return -mVal;
            return val;
        }

        public bool Store(JSONClass jc, bool forceStore = false)
        {
            string storeName = region;
            if (sideId == 1) return false;
            if (sideId == 0) storeName = region.Substring(1);
            JSONClass tempJC = new JSONClass();
            bool needsStore = strengthJSON.Store(tempJC, forceStore);
            if (strengthTenseJSON.Store(tempJC, forceStore)) needsStore = true;
            if (frequencyJSON.Store(tempJC, forceStore)) needsStore = true;
            if (frequencyTenseJSON.Store(tempJC, forceStore)) needsStore = true;
            if (dampingJSON.Store(tempJC, forceStore)) needsStore = true;
            if (dampingTenseJSON.Store(tempJC, forceStore)) needsStore = true;
            if (minAccelJSON.Store(tempJC, forceStore)) needsStore = true;
            if (maxAccelJSON.Store(tempJC, forceStore)) needsStore = true;
            if (maxMorphJSON.Store(tempJC, forceStore)) needsStore = true;
            if (needsStore) jc[storeName] = tempJC;
            return needsStore;
        }

        public void Load(JSONClass jc)
        {
            string loadName = region;
            if (sideId >= 0) loadName = region.Substring(1);
            if (jc.HasKey(loadName))
            {
                jc = jc[loadName].AsObject;
                maxMorphJSON.Load(jc);
                strengthJSON.Load(jc);
                frequencyJSON.Load(jc);
                dampingJSON.Load(jc);
                minAccelJSON.Load(jc);
            }
        }
    }
}