using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Leap.Unity;
using SimpleJSON;
using UnityEngine;
using Object = UnityEngine.Object;

namespace CheesyFX
{
    public class JiggleRegion : MonoBehaviour
    {
        private new string name;
        private int id;
        private MVRScript script;
        private Rigidbody rb;
        private List<RBDynamics> rbDynamics = new List<RBDynamics>();
        private List<JiggleHandler> subHandlers = new List<JiggleHandler>();
        private List<JiggleHandler> combinedHandlers = new List<JiggleHandler>();
        public List<JiggleHandler> handlers = new List<JiggleHandler>();

        public int selectedSubregion;

        public JSONStorableBool isCombined = new JSONStorableBool("isCombined", true);
        public JSONStorableBool tensionEnabled = new JSONStorableBool("Tension enabled", true);

        public void Init(MVRScript script, int id, string name, string[] rbNames)
        {
            this.name = name;
            this.id = id;
            this.script = script;

            isCombined.setCallbackFunction += val =>
            {
                handlers.ForEach(x => x.Reset());
                handlers = val ? combinedHandlers : subHandlers;
            };
            tensionEnabled.setCallbackFunction += ToggleTension;
            
            string[] sides = { "l", "r" };
            for (int i=0; i< rbNames.Length; i++)
            {
                string regionName = name;
                if (rbNames.Length == 2) regionName = sides[i] + regionName;
                rb = script.containingAtom.forceReceivers.First(x => x.name == rbNames[i]).GetComponent<Rigidbody>();
                var rbDynamic = rb.AddRbDynamic();
                rbDynamic.SetTensionFunction();
                rbDynamics.Add(rbDynamic);
                SetupHandlers(regionName, rbDynamic);
            }
            handlers = combinedHandlers;
            
            if (id == 4) tensionEnabled.val = false;
        }

        private void SetupHandlers(string regionName, RBDynamics rbDynamic)
        {
            subHandlers.Add(new JiggleHandler(id, $"{regionName}Top", rbDynamic));
            subHandlers.Add(new JiggleHandler(id, $"{regionName}Mid", rbDynamic));
            subHandlers.Add(new JiggleHandler(id, $"{regionName}Low", rbDynamic));
            
            combinedHandlers.Add(new JiggleHandler(id, $"{regionName}", rbDynamic));
        }

        private void ToggleTension(bool val)
        {
            for (int i = 0; i < subHandlers.Count; i++) subHandlers[i].tensionEnabled = val;
            for (int i = 0; i < combinedHandlers.Count; i++) combinedHandlers[i].tensionEnabled = val;
        }

        public void FixedUpdate()
        {
            for (int i = 0; i < handlers.Count; i++) handlers[i].Jiggle();
        }

        public void Reset()
        {
            for (int i = 0; i < handlers.Count; i++) handlers[i].Reset();
        }

        public void OnDestroy()
        {
            rbDynamics.ForEach(x =>
            {
                x.uses--;
                if(x.uses == 0) Destroy(x);
            });
            Reset();
        }
        
        public void OnDisable()
        {
            Reset();
        }

        public bool Store(JSONClass jc, bool forceStore = false)
        {
            JSONClass tempJC = new JSONClass();
            var needsStore = isCombined.Store(tempJC, forceStore);
            if (tensionEnabled.Store(tempJC, forceStore) && !needsStore) needsStore = true;
            for (int i = 0; i < combinedHandlers.Count; i++)
            {
                if (combinedHandlers[i].Store(tempJC, forceStore) && !needsStore) needsStore = true;
            }
            for (int i = 0; i < subHandlers.Count; i++)
            {
                if (subHandlers[i].Store(tempJC, forceStore) && !needsStore) needsStore = true;
            }
            if (needsStore) jc[name] = tempJC;
            return needsStore;
        }
        
        public void Load(JSONClass jc)
        {
            if (jc.HasKey(name))
            {
                jc = jc[name].AsObject;
                isCombined.Load(jc);
                tensionEnabled.Load(jc);
                for (int i = 0; i < combinedHandlers.Count; i++)
                {
                    combinedHandlers[i].Load(jc);
                }
                for (int i = 0; i < subHandlers.Count; i++)
                {
                    subHandlers[i].Load(jc);
                }
            }
        }
    }
}