using System;
using UnityEngine;

namespace CheesyFX
{
    public class RBDynamics : MonoBehaviour
    {
        private int regionId;
        private int sideId;
        public Rigidbody rigidbody;
        private RBDynamics parent;
        private RBDynamics child;
        public Vector3 velocity;
        public Vector3 angularVelocity;
        private Vector3 _lastVelocity;
        private Vector3 _lastAngularVel;

        private Vector3 _acceleration;
        private bool _accelerationSet;

        private Vector3 _localAcceleration;
        private bool _localAccelerationSet;
        
        private Vector3 _angularAcceleration;
        private bool _angularAccelerationSet;
        
        private Vector3 _localAngularAcceleration;
        private bool _localAngularAccelerationSet;

        private float _tension;
        private bool _tensionSet;
        
        private float _tensionAlt;
        private bool _tensionAltSet;

        private float _activity;
        private bool _activitySet;

        public int uses;

        private Func<float> getTension;
        private Func<float> getTensionAlt;

        public void Init(Rigidbody rb, int regionId = -1, int sideId = -1)
        {
            this.regionId = regionId;
            this.sideId = sideId;
            rigidbody = rb;
            getTension = () => rigidbody.transform.localRotation.x;
            getTensionAlt = () => rigidbody.transform.localRotation.z;
        }

        public void FixedUpdate()
        {
            _lastVelocity = velocity;
            _lastAngularVel = angularVelocity;
            velocity = rigidbody.velocity;
            angularVelocity = rigidbody.angularVelocity;
            _accelerationSet = false;
            _localAccelerationSet = false;
            _angularAccelerationSet = false;
            _localAngularAccelerationSet = false;
            _activitySet = false;
            _tensionSet = false;
            _tensionAltSet = false;
        }

        public Vector3 localAcceleration
        {
            get
            {
                if (_localAccelerationSet) return _localAcceleration;
                _localAcceleration = rigidbody.transform.InverseTransformDirection(acceleration);
                _localAccelerationSet = true;
                return _localAcceleration;
            }
        }
        
        public Vector3 acceleration
        {
            get
            {
                if (_accelerationSet) return _acceleration;
                _acceleration = (velocity - _lastVelocity) / Time.fixedUnscaledDeltaTime;
                _accelerationSet = true;
                return _acceleration;
            }
        }

        public Vector3 angularAcceleration
        {
            get
            {
                if (_angularAccelerationSet) return _angularAcceleration;
                _angularAcceleration = (angularVelocity - _lastAngularVel) / Time.fixedUnscaledDeltaTime;
                _angularAccelerationSet = true;
                return _angularAcceleration;
            }
        }

        public Vector3 localAngularAcceleration
        {
            get
            {
                if (_localAngularAccelerationSet) return _localAngularAcceleration;
                _localAngularAcceleration = rigidbody.transform.InverseTransformDirection(angularAcceleration);
                _localAngularAccelerationSet = true;
                return _localAngularAcceleration;
            }
        }

        public float tension
        {
            get
            {
                if (_tensionSet) return _tension;
                _tension = getTension();
                _tensionSet = true;
                return _tension;
            }
        }
        public float tensionAlt
        {
            get
            {
                if (_tensionAltSet) return _tensionAlt;
                // _tension = 1f-1.5f*Mathf.Abs(rigidbody.transform.localRotation.x);
                _tensionAlt = getTensionAlt();
                _tensionAltSet = true;
                return _tensionAlt;
            }
        }
        
        public float activity
        {
            get
            {
                if (_activitySet) return _activity;
                float act = rigidbody.mass * Math.Abs(Vector3.Dot(acceleration, velocity));
                for (int i = 0; i < 3; i++)
                {
                    act += Mathf.Abs(angularAcceleration[i] * rigidbody.inertiaTensor[i] * angularVelocity[i]);
                }
                _activity = act;
                _activitySet = true;
                return act;
            }
        }

        private void GetRelatives()
        {
            parent = rigidbody.transform.parent.GetComponent<Rigidbody>().AddRbDynamic();
            for (int i = 0; i < rigidbody.transform.childCount; i++)
            {
                var rb = rigidbody.transform.GetChild(i).GetComponent<Rigidbody>();
                if (rb != null)
                {
                    child = rb.AddRbDynamic();
                    break;
                }
            }
            // (rigidbody.name+":"+parent.name+":"+child.name).Print();
        }

        public void SetTensionFunction()
        {
            GetRelatives();
            getTension = () => 1f - Mathf.Clamp(1.2f * Mathf.Abs(rigidbody.transform.localRotation.x), 0f, 1f);
            if (rigidbody.name.Contains("Shin"))
            {
                getTension = () => 1f - Mathf.Clamp(1.2f * Mathf.Abs(rigidbody.transform.localRotation.x), 0f, 1f);
            }
            else if (rigidbody.name.Contains("abdomen"))
            {
                getTension = () => Mathf.Clamp(.3f + parent.tension - rigidbody.transform.localRotation.x, 0f, 1f);
            }
            else if(rigidbody.name.Contains("Thigh")) getTension = () =>
            {
                float thisTension = Mathf.Clamp(1.2f * Mathf.Abs(rigidbody.transform.localRotation.x), 0f, 1f);
                float childTension = child.tension;
                return (thisTension + childTension) * .5f;
            };
        }
    }

    public static class Extensions
    {
        public static RBDynamics AddRbDynamic(this Rigidbody rb)
        {
            var rbDynamic = rb.gameObject.GetComponent<RBDynamics>();
            if (rbDynamic == null)
            {
                rbDynamic = rb.gameObject.AddComponent<RBDynamics>();
                rbDynamic.Init(rb);
            }
            rbDynamic.uses ++;
            return rbDynamic;
        }
    }
}